from math import *
import string
import re
import os.path
import datetime
import qubtree
import QUB
from Buttons import AddTaskButton
from Dialogs import *

try:
    from qubopt import BalanceAllLoops
except:
    def BalanceAllLoops(mdl):
        QUB.Report("HyperQub requires a 2006 QUB for detailed balance constraints")
        return mdl
    
def concat(lsts):
    return reduce(lambda x,y:x+y, lsts)

def cartprod(*lsts):
    def extend(prod, remains):
        if 0 == len(remains):
            return [prod]
        subremains = remains[1:]
        return concat([extend(prod+[el], subremains) for el in remains[0]])
    return extend([], lsts)

class HQState:
    def __init__(self, i, x, y, cls, pr):
        self.i = i
        self.x = x
        self.y = y
        self.cls = cls
        self.pr = pr
        self.outs = {} # dim -> HQRate

class HQRate:
    def __init__(self, sfrom, sto, dim, k0fw, k0bk):
        self.sfrom = sfrom
        self.sto = sto
        self.dim = dim
        self.k0fw = k0fw
        self.k0bk = k0bk

class HQModel:
    def __init__(self):
        self.states = []
        self.rates = []
        self.constraints = []
        self.nchannel = 1
    def addRate(self, sifrom, sito, dim, k0fw, k0bk):
        sfrom = self.states[sifrom]
        sto = self.states[sito]
        rate = HQRate(sfrom, sto, dim, k0fw, k0bk)
        sfrom.outs[dim] = rate
        sto.outs[dim] = rate
        self.rates.append(rate)
    def delRate(self, rate):
        del rate.sfrom.outs[rate.dim]
        del rate.sto.outs[rate.dim]
        self.rates.remove(rate)
    def rateOfStates(self, sifrom, sito):
        sfrom = self.states[sifrom]
        sto = self.states[sito]
        for dim, rate in sfrom.outs.items():
            if sto == rate.sto:
                return rate
        return None

def TestLinear(mdl):
    hmdl = HQModel()
    state = mdl.find('States').find('State')
    i = 0
    while not state.isNull:
        hmdl.states.append(HQState(i, state['x'].data[0], state['y'].data[0], state['Class'].data[0], state['Pr'].data[0]))
        state = state.nextSameName()
        i += 1
    rate = mdl.find('Rates').find('Rate')
    i = 1
    while not rate.isNull:
        sts = rate['States'].data
        k0s = rate['k0'].data
        hmdl.addRate(sts[0], sts[1], i, k0s[0], k0s[1])
        rate = rate.nextSameName()
        i += 1 # each in own dimension so they don't overwrite
    nstate = len(hmdl.states)
    at = None
    for state in hmdl.states:
        if len(state.outs) == 1:
            at = state
            break
    if at == None:
        return False, 0, [], []
    classChangeAfter = []
    linRates = []
    visited = {}
    cls = at.cls
    i = 0
    while True:
        if visited.has_key(at):
            return False, 0, [], []
        visited[at] = True
        if not len(at.outs):
            break
        rate = at.outs.values()[0]
        if at == rate.sfrom:
            at = rate.sto
            linRates.append((rate.k0fw, rate.k0bk))
        else:
            at = rate.sfrom
            linRates.append((rate.k0bk, rate.k0fw))
        i += 1
        if cls != at.cls:
            classChangeAfter.append(i)
        cls = at.cls
        hmdl.delRate(rate)
    isLinear = (len(visited) == nstate) and (len(hmdl.rates) == 0)
    return isLinear, nstate-1, classChangeAfter, linRates

class HQub:
    def __init__(self, n=0, classChangeAfter=[], defOnPath=100.0, defOffPath=0.1, baseModel=None):
        if baseModel:
            isLinear, linN, linCCA, linRates = TestLinear(baseModel)
            if isLinear:
                self.Build(linN, linCCA, defOnPath, defOffPath)
                # copy on-path rates
                state = self.mdl.states[0]
                dim = 1
                while dim <= linN:
                    rate = state.outs[dim]
                    group = self.SameNeighborGroup(state, rate.sto)
                    for pair in group:
                        rate = self.mdl.rateOfStates(pair[0], pair[1])
                        rate.k0fw, rate.k0bk = linRates[dim-1]
                    state = rate.sto
                    dim += 1
            else: # init with nonlinear, assume it's one of our hypercubes
                nstate = 0
                state = baseModel.find('States').find('State')
                while not state.isNull:
                    nstate += 1
                    state = state.nextSameName()
                n = int(round(ceil(log(nstate) / log(2)))) # smallest n with at least nstate
                self.Build(n, classChangeAfter, defOnPath, defOffPath)
                #copy states
                i = 0
                state = baseModel.find('States').find('State')
                while not state.isNull:
                    hstate = self.mdl.states[i]
                    hstate.x = state['x'].data[0]
                    hstate.y = state['y'].data[0]
                    hstate.cls = state['Class'].data[0]
                    i += 1
                    state = state.nextSameName()
                #copy rates
                i = 0
                rate = baseModel.find('Rates').find('Rate')
                while (i < len(self.mdl.rates)) and not rate.isNull:
                    hrate = self.mdl.rates[i]
                    k0s = rate['k0'].data
                    hrate.k0fw = k0s[0]
                    hrate.k0bk = k0s[1]
                    i += 1
                    rate = rate.nextSameName()
                #copy constraints
                cns = baseModel.find('Constraints').child
                while not cns.isNull:
                    self.mdl.constraints.append([cns.name]+[x for x in cns.data])
                    cns = cns.sibling
            # final stuff for any kind of base model
            self.mdl.nchannel = baseModel['ChannelCount'].data[0]
        else: # no base model
            self.Build(n, classChangeAfter, defOnPath, defOffPath)
    def Build(self, n, classChangeAfter, defOnPath, defOffPath):
#        def SR(i):
#            if i == 0:
#                s0 = HQState(0, 0, 0, 0, 0)
#                s0.clsCanChange = True
#                return [s0], []
#            else:
#                lastS, lastR = SR(i-1)
#                S = []
#                classChange = i in classChangeAfter
#                for j, s in enumerate(lastS):
#                    sstay = HQState(2*s.i, 0, 0, s.cls, 0)
#                    sstay.clsCanChange = False
#                    S.append(sstay)
#                    if classChange and s.clsCanChange:
#                        cls = s.cls + 1
#                    else:
#                        cls = s.cls
#                    sgo = HQState(2*s.i+1, 0, 0, cls, 0)
#                    sgo.clsCanChange = s.clsCanChange
#                    S.append(sgo)
#                Rold = [HQRate(S[2*r.sfrom.i], S[2*r.sto.i], r.dim, r.k0fw, r.k0bk) for r in lastR]
#                Rnew = [HQRate(S[2*r.sfrom.i+1], S[2*r.sto.i+1], r.dim, r.k0fw, r.k0bk) for r in lastR]
#                Rboth = [HQRate(S[2*j], S[2*j+1], i, defOffPath, defOffPath) for j in range(len(lastS))]
#                R = Rold + Rnew + Rboth
#                return S, R
        def SR(i):
            if i == 0:
                s0 = HQState(0, 0, 0, 0, 0)
                return [s0], []
            else:
                lastS, lastR = SR(i-1)
                S = []
                classChange = i in classChangeAfter
                for j, s in enumerate(lastS):
                    sstay = HQState(2*s.i, 0, 0, s.cls, 0)
                    S.append(sstay)
                    if classChange:
                        cls = i
                    else:
                        cls = s.cls
                    sgo = HQState(2*s.i+1, 0, 0, cls, 0)
                    S.append(sgo)
                Rold = [HQRate(S[2*r.sfrom.i], S[2*r.sto.i], r.dim, r.k0fw, r.k0bk) for r in lastR]
                Rnew = [HQRate(S[2*r.sfrom.i+1], S[2*r.sto.i+1], r.dim, r.k0fw, r.k0bk) for r in lastR]
                Rboth = [HQRate(S[2*j], S[2*j+1], i, defOffPath, defOffPath) for j in range(len(lastS))]
                R = Rold + Rnew + Rboth
                return S, R
        S, R = SR(n)
        self.n = n
        self.mdl = HQModel()
        self.mdl.states = S
        for r in R:
            self.mdl.addRate(r.sfrom.i, r.sto.i, r.dim, r.k0fw, r.k0bk)
        self.SetOnPathRates(defOnPath)
    def OnPathPairs(self):
        def stateName(conf):
            return string.join(conf, '')
        def stateIx(conf):
            return self.stateNameToIx(stateName(conf))
        onpath = []
        at = ['-'] * self.n
        for d in range(self.n):
            next = at[:]
            next[d] = '+'
            onpath.append( [stateIx(at), stateIx(next)] )
            at = next
        return onpath
    def SameNeighborGroups(self):
        def listconf(n, symbols=['-', '+']):
            return apply(cartprod, n*[symbols])
        def stateName(conf):
            return string.join(conf, '')
        def stateIx(conf):
            return self.stateNameToIx(stateName(conf))
        def mkconf(d, dconf, near, nconf, far, fconf):
            conf = [0]*(1+len(near)+len(far))
            conf[d-1] = dconf
            for lst, lconf in [(near, nconf), (far, fconf)]:
                for i in range(len(lst)):
                    conf[lst[i]-1] = lconf[i]
            return conf
        for d in range(1, self.n+1):
            near = [i for i in [d-1, d+1] if 0<i<=self.n]
            far  = [i for i in range(1, self.n+1) if not (i==d or (i in near))]
            nearconf = listconf(len(near))
            farconf  = listconf(len(far))
            for nconf in nearconf:
                statePairs = [ [stateIx(mkconf(d,sym,near,nconf,far,fconf)) for sym in ['-','+']]
                                for fconf in farconf ]
                yield statePairs
    def SameNeighborGroup(self, s1, s2):
        def indexOf(lst, x):
            for i in range(len(lst)):
                if lst[i] == x:
                    return i
            return -1
        pair = [indexOf(self.mdl.states, s) for s in [s1, s2]]
        for group in self.SameNeighborGroups():
            if pair in group:
                return group
        return []
    def SetOnPathRates(self, defOnPath):
        onpath = self.OnPathPairs()
        for statePairs in self.SameNeighborGroups():
            statePairsOffPath = [pair for pair in statePairs if not pair in onpath]
            if len(statePairsOffPath) < len(statePairs):
                for pair in statePairs:
                    rate = self.mdl.rateOfStates(pair[0], pair[1])
                    rate.k0fw, rate.k0bk = defOnPath, defOnPath
    def AsTree(self):
        tree = qubtree.Node('ModelFile')
        tree['ChannelCount'].data = self.mdl.nchannel
        states = tree['States']
        node = qubtree.NullNode()
        for state in self.mdl.states:
            node = states.insert('State', node)
            node['x'].data = state.x
            node['y'].data = state.y
            node['Class'].data = state.cls
            node['Pr'].data = state.pr
        rates = tree['Rates']
        node = qubtree.NullNode()
        for rate in self.mdl.rates:
            node = rates.insert('Rate', node)
            node['States'].data = [state.i for state in [rate.sfrom, rate.sto]]
            node['k0'].data = [rate.k0fw, rate.k0bk]
            node['k1'].data = [0.0, 0.0]
            node['P'].data = [0, 0]
            node['Q'].data = [0, 0]
        constraints = tree['Constraints']
        node = qubtree.NullNode()
        for cns in self.mdl.constraints:
            node = constraints.insert(cns[0], node)
            node.data = cns[1:]
        return tree
    def AddFixPathConstraints(self):
        onpath = self.OnPathPairs()
        for statePairs in self.SameNeighborGroups():
            statePairsOffPath = [pair for pair in statePairs if not pair in onpath]
            if len(statePairsOffPath) < len(statePairs): # if this group includes an on-path edge
                for a, b in statePairs:
                    self.mdl.constraints.append( ['FixRate', a, b] )
                    self.mdl.constraints.append( ['FixRate', b, a] )
    def AddScaleConstraints(self):
        onpath = self.OnPathPairs()
        for statePairs in self.SameNeighborGroups():
            statePairsOffPath = [pair for pair in statePairs if not pair in onpath]
            if len(statePairs) == len(statePairsOffPath): # different neighbor conditions from any on-path edge
                self.mdl.constraints += [ ['ScaleRate'] + statePairs[0] + pair for pair in statePairs[1:] ]
                for pair in statePairs:
                    pair.reverse()
                self.mdl.constraints += [ ['ScaleRate'] + statePairs[0] + pair for pair in statePairs[1:] ]
    def AddLoopConstraints(self):
        mdl = self.AsTree()
        mdl['Amps'].data = map(float, range(10))
        mdl['Stds'].data = [0.1] * 10
        mdl['NAr'].data = [0] * 10
        #print datetime.datetime.now()
        balanced = BalanceAllLoops(mdl)
        #print datetime.datetime.now()
        loop = balanced.find('Constraints').find('LoopBal')
        while not loop.isNull:
            self.mdl.constraints.append(['LoopBal']+[x for x in loop.data])
            loop = loop.nextSameName()
    def stateNameToIx(self, name):
        def nondash(c):
            if c == '-':
                return 0
            else:
                return 1
        def dimval(i, c):
            ii = self.n - i - 1
            return int(round(pow(2, ii))) * nondash(c)
        return sum(map(lambda ic: dimval(ic[0], ic[1]), enumerate(name)))
    def stateIxToName(self, ix):
        name = []
        for i, p in [(j, int(round(pow(2, self.n - j - 1)))) for j in range(self.n)]:
            if ix & p:
                name.append(self.dimIxToName(i+1))
            else:
                name.append('-')
        return string.join(name, '')
    def dimIxToName(self, ix):
        return chr( ord('A') + ix - 1 )
    def dimNameToIx(self, name):
        return ord(name.upper()) - ord('A') + 1

class HyperQuber:
    def __init__(self):
        self.model = QUB.ModelView.currModel
        if not self.model:
            QUB.Report('HyperQub: no model available.')
            return
        
        baseModel = self.model.getAsNode()
        self.InitModel(baseModel)

        self.wiz = QUB.CreateWizard()
        self.wiz.caption = 'HyperQub: %s' % os.path.split(str(baseModel['FileName'].data))[1]
        self.wiz.buttons = ['Build', 'Constrain', 'Lay out', 'Done']
        self.wiz.show()
        self.layoutmsg = ''

        self.properties = QUB.Profiles['HyperQub'].active
        def initProp(name, val):
            if self.properties.find(name).isNull:
                self.properties[name].data = val
        initProp('initLinear', 0)
        initProp('n', 3)
        initProp('cca', '')
        initProp('defOn', 100.0)
        initProp('defOff', 0.1)
        initProp('fixOnPath', 1)
        initProp('scaleGroups', 1)
        initProp('balLoops', 1)
        initProp('order', '')
        initProp('feature', '')
    def InitModel(self, baseModel):
        self.hqub = HQub(baseModel=baseModel)
    def ReadButtons(self):
        btn = None
        while btn != 'Done':
            self.wiz.message = ('%d dimensional hypercube\r\n%s\r\n%d constraints | %d free' %
                                (self.hqub.n, self.layoutmsg, len(self.hqub.mdl.constraints),
                                 (2*len(self.hqub.mdl.rates)-len(self.hqub.mdl.constraints))))
            btn = self.wiz.waitForButton()
            if btn == 'Build':
                self.PromptBuildModel()
            elif btn == 'Constrain':
                self.PromptConstrainModel()
            elif btn == 'Lay out':
                self.PromptLayout()
    def PromptBuildModel(self):
        def acceptDims(s):
            dims = map(self.hqub.dimNameToIx, s)
            return dims
        def formatDims(dims):
            return string.join(map(self.hqub.dimIxToName, dims), '')
        props = self.properties
        response = QUB.Dialog('Build HyperQub', ['OK', 'Cancel'],
                              [DialogCheckBox('initLinear', 'Initialize using existing linear model', props['initLinear'].data[0]),
                               DialogText('n', '# blocks/dimensions:', props['n'].data[0], acceptIntBetween(1,7)),
                               DialogText('cca', 'Class change on dimension(s) e.g. C:', acceptDims(str(props['cca'].data)), acceptDims, formatDims),
                               DialogText('defOn', 'Default on-path rate:', props['defOn'].data[0], acceptFloatGreaterThan(0.0)),
                               DialogText('defOff', 'Default off-path rate:', props['defOff'].data[0], acceptFloatGreaterThan(0.0))])
        if response['button'] == 'OK':
            if response['initLinear']:
                self.hqub = HQub(response['n'], response['cca'], response['defOn'], response['defOff'], self.model.getAsNode())
            else:
                self.hqub = HQub(response['n'], response['cca'], response['defOn'], response['defOff'])
            self.Layout()
            self.PutModel()
        props['initLinear'].data = int(response['initLinear'])
        props['n'].data = response['n']
        props['cca'].data = formatDims(response['cca'])
        props['defOn'].data = response['defOn']
        props['defOff'].data = response['defOff']
    def PromptConstrainModel(self):
        props = self.properties
        response = QUB.Dialog('Constrain HyperQub', ['OK', 'Cancel'],
                              [DialogCheckBox('fixOnPath', 'Fix on-path rates', props['fixOnPath'].data[0]),
                               DialogCheckBox('scaleGroups', 'Scale rates with same neighbors', props['scaleGroups'].data[0]),
                               DialogCheckBox('balLoops', 'Detailed balance', props['balLoops'].data[0])])
        if response['button'] == 'OK':
            self.hqub.mdl.constraints[:] = []
            if response['fixOnPath']:
                self.hqub.AddFixPathConstraints()
            if response['scaleGroups']:
                self.hqub.AddScaleConstraints()
            if response['balLoops']:
                self.wiz.message = 'Paring superfluous constraints...\r\nthis could take a very long time\r\n(make sure to save right away!)'
                self.hqub.AddLoopConstraints()
            self.PutModel()
        props['fixOnPath'].data = int(response['fixOnPath'])
        props['scaleGroups'].data = int(response['scaleGroups'])
        props['balLoops'].data = int(response['balLoops'])
    def PromptLayout(self):
        def acceptOrder(s):
            order = map(self.hqub.dimNameToIx, s)
            for i in order:
                if i <= 0 or i > self.hqub.n:
                    raise FieldAcceptException(self.hqub.dimIxToName(i), 'invalid dimension letter')
            return order
        def formatOrder(order):
            return string.join(map(self.hqub.dimIxToName, order), '')
        def acceptStates(s):
            toks = re.split(',? *', s)
            if len(toks) == 1 and toks[0] == '':
                toks = []
            ixs = map(self.hqub.stateNameToIx, toks)
            for i in range(len(ixs)):
                if ixs[i] < 0 or ixs[i] >= len(self.hqub.mdl.states):
                    raise FieldAcceptException(toks[i], 'invalid state name')
            return ixs
        def formatStates(ixs):
            return string.join(map(self.hqub.stateIxToName, ixs), ', ')
        props = self.properties
        response = QUB.Dialog('Lay out HyperQub', ['OK', 'Cancel'],
                              [DialogText('order', 'dimension order e.g. ABCD:', acceptOrder(str(props['order'].data)), acceptOrder, formatOrder),
                               DialogText('feature', 'feature states e.g. -B--:', acceptStates(str(props['feature'].data)), acceptStates, formatStates)])
        if response['button'] == 'OK':
            self.Layout(response['order'], response['feature'])
            self.PutModel()
        props['order'].data = formatOrder(response['order'])
        props['feature'].data = formatStates(response['feature'])
    def Layout(self, order=[], feature=[]):
        mx, my = 15.0, 15.0
        w = 100.0 - 2*mx
        yfeat = 15.0
        if 0 == len(feature):
            yfeat = my/2
        elif len(feature) > 1:
            x0feat = mx
            dfeat = w/(len(feature) - 1)
        else:
            x0feat = 50.0
            dfeat = 0.0
        y0 = 2*yfeat
        h = 100.0 - y0 - 2*my
        offset = [(45.0, 0.0), (0.0, 45.0), (14.0, 16.0), (-5.0, 8.0), (10.0, -5.0), (8.0, 5.0), (5.0, 8.0)]
        present = {}
        for i in order:
            present[i] = True
        fullorder = order[:]
        for i in range(1, self.hqub.n+1):
            if not present.has_key(i):
                fullorder.append(i)

        self.layoutmsg = ('dimension order: %s\r\nfeatured: %s' %
                          (string.join(map(self.hqub.dimIxToName, fullorder), ''),
                           string.join(map(self.hqub.stateIxToName, feature), ', ')))

        hmdl = self.hqub.mdl
        states = hmdl.states
        state = states[0]
        state.x, state.y = mx, y0
        vis = [state]
        for d, dim in enumerate(fullorder):
            xoff, yoff = offset[d]
            for state in vis[:]:
               newstate = state.outs[dim].sto
               newstate.x = state.x + xoff
               newstate.y = state.y + yoff
               vis.append(newstate)
        for ix in feature:
            states[ix].x, states[ix].y = x0feat, yfeat
            x0feat += dfeat
    def PutModel(self):
        self.wiz.message = 'Drawing model ... please wait'
        self.model.setAsNode(self.hqub.AsTree())

def HyperQub(props):
    try:
        HyperQuber().ReadButtons()
    except Exception, e:
        QUB.Report(e)


AddTaskButton( 'HyperQub', # button text
               HyperQub,   # function
               'HQ',       # short text for small buttons
               0,            # show settings dialog on left-click?
               0,            # show settings dialog on right-click?
               "Hypercube model building and visualization",   # tool-tip
               "Hypercube model building and visualization",   # description in "Edit Buttons"
               "Properties:", # caption for the settings dialog
               [],
               'Modeling')



